#pragma once
#include <cv.h>   //   OpenCV.
#include <vector> //    std::vector.

//  ,   
//   .
class CPixelMappingModel
{
public:
  
  //   .
  virtual int GetParamCount() = 0;

  //    pDstPixel,   
  //   pSrcPixel,     
  //  .
  //  true   , false   .
  virtual bool MapPixel(
     CvPoint2D32f* pSrcPixel, // [in]   .          
     CvPoint2D32f* pDstPixel  // [out]  , . .
    ) = 0;

  //        .
  //  true   , false   .
  virtual bool Reset() = 0;

  //      .
  //  true   , false   .
  virtual bool Update(
    CvMat* pMatParams  // [in]  .
   ) = 0;

  //        .
  //  true   , false   .
  virtual bool CalcJacobianMat(
    CvMat* pMatJacobian, // [out]  .
    CvPoint2D32f* pPixel // [in]  .
   ) = 0;  
};

//   .
enum CRITERIA_TYPE 
{
  NONE = 0,              //     .
  ACCURACY_REACHED = 1,  //   .
  MAX_ITERS_REACHED = 2, //  .  .
  OSCILLATION = 3        //    .
};

//  .
struct CriteriaContext
{
  CriteriaContext()
  {
    m_nBadIters = 0;
    m_fMinObjFuncVal = -1;
  }

  int m_nBadIters;        // -  ,    .
  float m_fMinObjFuncVal; //     .
};

// ,  -  
//  .
class CInvCompRegistrationAlgorithm 
{
  friend class CTest;
public:

  // . 
  CInvCompRegistrationAlgorithm();

  // .
  ~CInvCompRegistrationAlgorithm(); 

  //      I1  I2.
  bool RegisterImages(
    IplImage* pImgI1,      // [in]  I1.
    IplImage* pImgConfMap, // [in]    I1.
    IplImage* pImgI2,      // [in]  I2.
    CPixelMappingModel* pPixelMapping, // [in]   .
    float fAccuracy = 10E-03, // [in]     .
    int nMaxIters = 100,      // [in]   .
    int nMaxOscillatingIters = 5, // [in] .  ,    .
    int* pnActualIters = NULL,// [out]   .
    float fInnerAccuracy = 10E-03, // [in]      .
    int nInnerMaxIters = 1    // [in]       .
   );

  //  .   ( )  , 
  //    ,   .   
  //  true   ,    - false.
  bool AllocResources(
    CvSize ImageSize, // [in]    .
    CPixelMappingModel* pPixelMapping // [in]   .
   );

  //    ,  
  //  .
  void FreeResources();

  //      
  //  ,  " " 
  //       .
  bool Precompute(
    IplImage* pImgI1,      // [in]  I1.
    IplImage* pImgConfMap, // [in]    I1.
    CPixelMappingModel* pPixelMapping // [in]   .
   );

  //       .
  //         , 
  //      .
  float ApproximateParameterIncrement(        
        IplImage* pImgI1,      // [in]  I1.
        IplImage* pImgConfMap, // [in]     I1.        
        CPixelMappingModel* pPixelMapping, // [in]   .
        CvMat* pMatPrevParams, // [in]     .
        CvMat* pMatNewParams   // [out]     .        
      );
  
  //        
  //  .   , 
  //       .
  CRITERIA_TYPE IsInnerLoopReady(
    CvMat* pMatPrevParams, // [in]     .
    CvMat* pMatNewParams,  // [in]     .
    float fAccuracy,  // [in]  .
    int nCurrentIter, // [in]   ,   .
    int nMaxIters     // [in]   .
   );

  //  ()  I2   
  //      ,  
  // ""    I1.
  IplImage* WarpImageI2(
    IplImage* pImgI2,      // [in]  I2.
    IplImage* pImgConfMap, // [in]     I1.    
    CPixelMappingModel* pPixelMapping // [in]   .    
   );

  //     .   ,  
  //      .
  CRITERIA_TYPE IsReady(
    CvMat* pMatParams, // [in]  .
    float fAccuracy,   // [in]  .
    int nCurrentIter,  // [in]   ,   .
    int nMaxIters,     // [in]    .
    float fObjFuncVal, // [in]    .    
    int nMaxOscillatingIters, // [in]    .
    CriteriaContext& context  // [in]  .
   );

private:

  //    (x;y), 
  //   .
  float InterpolatePixelIntensity(
    IplImage* pImg, // [in] .
    float x,        // [in]   x.
    float y         // [in]   y.
   );

  /* ,   . */
  
  //    
  //    I2.
  IplImage* m_pImgWarpedI2;      
                                 
  // ,   I1   X. 
  IplImage* m_pImgGradI1x;        
                                 
  // ,   I1   Y. 
  IplImage* m_pImgGradI1y;        
  
  //   " ".
  std::vector<IplImage*> m_apImgSteepestDescent; 
                              
  //     m x 1 ().
  CvMat* m_pMatGradF;    

  //        m x m.
  CvMat* m_pMatInverseHessianF;  
  
  //   m x m. 
  CvMat* m_pMatA;                
  
  //   m x 1. 
  CvMat* m_pMatV;   

  //      m x 1.
  CvMat* m_pMatPrevParams;

  //      m x 1.
  CvMat* m_pMatNewParams;
};

