#include "PixelMapping3DOF.h"
#define _USE_MATH_DEFINES
#include <math.h>

CPixelMappingModel3DOF::CPixelMappingModel3DOF()
{
  //   .
  m_pMatComposition = cvCreateMat(3, 3, CV_32FC1);
  m_pMatTwist = cvCreateMat(3,3, CV_32FC1);
  m_pMatA = cvCreateMat(3,3, CV_32FC1);

  //    .
  Reset();
}

CPixelMappingModel3DOF::~CPixelMappingModel3DOF()
{
  //  .
  cvReleaseMat(&m_pMatComposition);  
  cvReleaseMat(&m_pMatTwist);
  cvReleaseMat(&m_pMatA);
}

int CPixelMappingModel3DOF::GetParamCount()
{
  //     .
  return 3;
}

bool CPixelMappingModel3DOF::MapPixel(
   CvPoint2D32f* pSrcPixel, 
   CvPoint2D32f* pDstPixel  
  )
{
   //     .

    pDstPixel->x = 
      CV_MAT_ELEM(*m_pMatComposition, float, 0, 0)*pSrcPixel->x +   
      CV_MAT_ELEM(*m_pMatComposition, float, 0, 1)*pSrcPixel->y +
      CV_MAT_ELEM(*m_pMatComposition, float, 0, 2);

    pDstPixel->y = 
      CV_MAT_ELEM(*m_pMatComposition, float, 1, 0)*pSrcPixel->x +   
      CV_MAT_ELEM(*m_pMatComposition, float, 1, 1)*pSrcPixel->y +
      CV_MAT_ELEM(*m_pMatComposition, float, 1, 2);

  return true;
}

bool CPixelMappingModel3DOF::Reset()
{
  cvSetIdentity(m_pMatComposition);
  return true;
}

bool CPixelMappingModel3DOF::Update(
  CvMat* pMatParams       
 )
{
  //    Delta_p=(wz,tx,ty).
  float wz = CV_MAT_ELEM(*pMatParams, float, 0, 0);
  float tx = CV_MAT_ELEM(*pMatParams, float, 1, 0);
  float ty = CV_MAT_ELEM(*pMatParams, float, 2, 0);

  //  -.
  cvSetIdentity(m_pMatTwist);
  CV_MAT_ELEM(*m_pMatTwist, float, 0, 1) = -wz;
  CV_MAT_ELEM(*m_pMatTwist, float, 1, 0) = wz;
  CV_MAT_ELEM(*m_pMatTwist, float, 0, 2) = tx;
  CV_MAT_ELEM(*m_pMatTwist, float, 1, 2) = ty;
  
  //   .
  double dDet = cvInvert(m_pMatTwist, m_pMatA);
  if(dDet==0)
  {
    return false;
  }

  //   .
  cvGEMM(m_pMatA, m_pMatComposition, 1, NULL, 0, m_pMatTwist);
  cvCopy(m_pMatTwist, m_pMatComposition);

  return true;
}

bool CPixelMappingModel3DOF::CalcJacobianMat(
  CvMat* pMatJacobian, 
  CvPoint2D32f* pPixel 
 )
{    
  CV_MAT_ELEM(*pMatJacobian, float, 0, 0) = -pPixel->y;
  CV_MAT_ELEM(*pMatJacobian, float, 1, 0) = pPixel->x;
  CV_MAT_ELEM(*pMatJacobian, float, 0, 1) = 1;
  CV_MAT_ELEM(*pMatJacobian, float, 1, 1) = 0;
  CV_MAT_ELEM(*pMatJacobian, float, 0, 2) = 0;
  CV_MAT_ELEM(*pMatJacobian, float, 1, 2) = 1;
  return true;
}

CvMat* CPixelMappingModel3DOF::GetComposition()
{
  return m_pMatComposition;
}

bool CPixelMappingModel3DOF::ExtractParams(
    CvPoint2D32f ptRotCenter, 
    float& fRotAngle,  
    float& fOffsX, 
    float& fOffsY  
   )
{ 
  //   .
  float r_21 = CV_MAT_ELEM(*m_pMatComposition, float, 1, 0);
  fRotAngle = -asin(r_21);
  float cos_a = cos(fRotAngle);
  float sin_a = sin(fRotAngle);
  fRotAngle *= 180.0f/(float)M_PI;
  
  //    .  
  fOffsX = CV_MAT_ELEM(*m_pMatComposition, float, 0, 2) - 
    (1-cos_a)*ptRotCenter.x + sin_a*ptRotCenter.y;
  fOffsY = CV_MAT_ELEM(*m_pMatComposition, float, 1, 2) -
    sin_a*ptRotCenter.x - (1-cos_a)*ptRotCenter.y;

  return true;
}

