#include "statfunc.h"
#include <math.h>
#include <cv.h>

CStatSample::CStatSample()
{ 
  m_bCalculated = false;
  m_fMinimum = 0;
  m_fMaximum = 0;
  m_fMean = 0;
  m_fVariance = 0;  
  m_fStdDev = 0;
}

CStatSample::~CStatSample()
{
}

void CStatSample::AddValue(double Xi)
{
  m_sample.push_back(Xi);
  m_bCalculated = false;
}

double CStatSample::GetValue(int i)
{
  assert(i<(int)m_sample.size());
  return m_sample[i];
}

size_t CStatSample::GetSize()
{
  return m_sample.size();
}

void CStatSample::CalcMoments()
{
  double N = (double)m_sample.size();

  if(N==0) return;
  
  // Calculate minimum and maximum values
  int i;
  for(i=0; i<N; i++)
  {
    if(i==0) 
    {
      m_fMinimum = m_fMaximum = m_sample[i];
    }
    else
    {
      if(m_sample[i]<m_fMinimum)
        m_fMinimum = m_sample[i];

      if(m_sample[i]>m_fMaximum)
        m_fMaximum = m_sample[i];
    }
  }

  // Calculate mean
  double sum = 0;  
  for(i=0; i<N; i++)
  {
    sum += m_sample[i];
  }

  m_fMean = sum/N;

  if(N<=1) return;

  // Calculate variance and stdandard deviation
  sum = 0;
  for(i=0; i<N; i++)
  {
    sum += (m_sample[i]-m_fMean)*(m_sample[i]-m_fMean);
  }

  m_fVariance = sum/N;
  m_fStdDev = sqrt(m_fVariance);
  m_fSampleStdDev = sqrt(sum/(N-1));

  m_bCalculated = true;    
}

double CStatSample::GetMinimumValue()
{
  assert(m_sample.size()>0);

  if(!m_bCalculated)
    CalcMoments();

  return m_fMinimum;
}
  
double CStatSample::GetMaximumValue()
{
  assert(m_sample.size()>0);

  if(!m_bCalculated)
    CalcMoments();

  return m_fMaximum;
}

double CStatSample::GetMeanValue()
{
  assert(m_sample.size()>0);
  
  if(!m_bCalculated)
    CalcMoments();

  return m_fMean;
}

double CStatSample::GetStdDev()
{
  assert(m_sample.size()>1);

  if(!m_bCalculated)
    CalcMoments();

  return m_fStdDev;
}

double CStatSample::GetSampleStdDev()
{
  assert(m_sample.size()>1);

  if(!m_bCalculated)
    CalcMoments();

  return m_fSampleStdDev;
}

void CStatSample::PrintStat(FILE* f, char* stat_name)
{  
  fprintf(f, "%s sample size (N) = %d\n", stat_name, GetSize());

  if(GetSize()>0)
  {
    fprintf(f, "%s minimum value  (min) = %g\n", stat_name, GetMinimumValue());
    fprintf(f, "%s maximum value  (max) = %g\n", stat_name, GetMaximumValue());
  }
  else
  {
    fprintf(f, "%s minimum value  (min) = n/a\n", stat_name);
    fprintf(f, "%s maximum value  (max) = n/a\n", stat_name);
  }

  if(GetSize()>0)
  {
    fprintf(f, "%s mean value  (E) = %g\n", stat_name, GetMeanValue());
  }
  else
  {
    fprintf(f, "%s mean value  (E) = n/a\n", stat_name);
  }

  if(GetSize()>1)
  {
    fprintf(f, "%s std. dev. (sigma) = %g\n", stat_name, GetStdDev());
    fprintf(f, "%s sample std. dev (s) = %g\n", stat_name, GetSampleStdDev());    
  }
  else
  {
    fprintf(f,"%s sample central moments are not available\n", stat_name);
  }  
}

