#include "Tests.h"

CTest::CTest()
{
  //  .
  m_pMatPrevParams = cvCreateMat(3, 1, CV_32FC1);
  m_pMatNewParams = cvCreateMat(3, 1, CV_32FC1);
}

CTest::~CTest()
{
  //  .
  cvReleaseMat(&m_pMatPrevParams);
  cvReleaseMat(&m_pMatNewParams);
}

  
void CTest::RunStepByStepDemo()
{
  //  -,       .
  FILE * f = NULL;
  _tfopen_s(&f, _T("out.txt"), _T("wt"));
  if(f==NULL)
  {
    //     .
    printf("Error: file out.txt can't be opened for writing.\n");
    return;
  }

  //  .
  cvNamedWindow("I1");
  cvMoveWindow("I1", 10, 200);
  cvNamedWindow("I2");
  cvMoveWindow("I2", 400, 200);
  cvNamedWindow("ConfMap");
  cvMoveWindow("ConfMap", 10, 500);
 
  //    .
  srand((unsigned)time(NULL));

  // C      .
  float fRotAngle = (float)( rand() % 10 - 10 );
  float fOffsX = (float)( rand() % 10 - 10);
  float fOffsY = (float)( rand() % 10 - 10);

  _tprintf(_T("RotAngle = %0.5g OffsX = %0.5g OffsY = %0.5g\n"), fRotAngle, fOffsX, fOffsY);
  _ftprintf(f, _T("RotAngle = %0.5g OffsX = %0.5g OffsY = %0.5g\n"), fRotAngle, fOffsX, fOffsY);

  //      .
  IplImage* pImgI1 = NULL;
  IplImage* pImgConfMap = NULL;
  IplImage* pImgI2 = NULL;

  //   .
  CvRect rcROI = cvRect(100, 50, 150, 150);    
  if(!CreateTestImages("Koala.jpg", &rcROI, 
    fRotAngle, fOffsX, fOffsY, &pImgI1, &pImgConfMap, &pImgI2))
  {
    //   .
    printf("Error creating test images.\n");
    return;
  }

  /*      I1  I2. */

  //  .
  CvSize ImageSize = cvSize(pImgI1->width, pImgI1->height);
  m_ImgRegAlg.AllocResources(ImageSize, &m_PixelMapping);

  //   .
  m_ImgRegAlg.Precompute(pImgI1, pImgConfMap, &m_PixelMapping);

  CriteriaContext context;

  //   .
  int l; //  .
  for(l=0; ; l++)
  {     
    //   I2.
    IplImage* pImgWarpedI2 = m_ImgRegAlg.WarpImageI2(pImgI2, pImgConfMap, &m_PixelMapping);
    
    //    .
    cvShowImage("I1", pImgI1);
    cvShowImage("I2", pImgWarpedI2);
    cvShowImage("ConfMap", pImgConfMap);
    cvWaitKey(100); //  100 .

    //        .
    cvSet(m_pMatPrevParams, cvScalar(0));

    //    .
    float F = 0;

    //   .
    int k;
    for(k = 0; ; k++)
    {        
      //      .
      F = m_ImgRegAlg.ApproximateParameterIncrement(
        pImgI1, pImgConfMap, &m_PixelMapping, m_pMatPrevParams, m_pMatNewParams);

      //     fRotAngle, fOffsX, fOffsY
      float fEstimatedRotAngle=0, fEstimatedOffsX=0, fEstimatedOffsY=0;
      CvPoint2D32f center = cvPoint2D32f(pImgI1->width/2, pImgI1->height/2);
      m_PixelMapping.ExtractParams(center, fEstimatedRotAngle, fEstimatedOffsX, fEstimatedOffsY);      

      //   ,   ,   ,
      //    fRotAngle, fOffsX, fOffsY
      _tprintf(_T("%d. F=%0.5g wz=%0.5g tx=%0.5g ty=%0.5g angle=%0.5g offsx=%0.5g offsy=%0.5g\n"),
        l,        
        F,
        CV_MAT_ELEM(*m_pMatNewParams, float, 0, 0),
        CV_MAT_ELEM(*m_pMatNewParams, float, 1, 0),
        CV_MAT_ELEM(*m_pMatNewParams, float, 2, 0),
        fEstimatedRotAngle,
        fEstimatedOffsX,
        fEstimatedOffsY );          

      //    -.
      _ftprintf(f, _T("%d %0.5g\n"), l, F);          

      //     .
      CRITERIA_TYPE InnerLoopReady = m_ImgRegAlg.IsInnerLoopReady(
        m_pMatPrevParams, m_pMatNewParams, 1E-03f, k, 1);
      
      //      .
      cvCopy(m_pMatNewParams, m_pMatPrevParams);        

      if(InnerLoopReady!=NONE)
        break;
    }
    
    //     .
    m_PixelMapping.Update(m_pMatNewParams);

    //      .
    CRITERIA_TYPE Ready = Ready=m_ImgRegAlg.IsReady(m_pMatNewParams, 1E-03f, l, 100, F, 5, context);
    if(Ready)
    {
      if(Ready==MAX_ITERS_REACHED) _tprintf(_T("Max. iters reached\n"));
      else if(Ready==ACCURACY_REACHED) _tprintf(_T("Accuracy reached\n"));
      else if(Ready==OSCILLATION) _tprintf(_T("Oscillations detected\n"));
      break;
    }
  }

  //     I1.
  CvPoint TopLeft = cvPoint(rcROI.x, rcROI.y);
  CvPoint BottomRight = cvPoint(rcROI.x+rcROI.width, rcROI.y+rcROI.height);
  cvRectangle(pImgI1, TopLeft, BottomRight, CV_RGB(255, 255,255));
  cvShowImage("I1", pImgI1);

  //     I2.
  DrawWarpedRect(pImgI2, rcROI, m_PixelMapping.GetComposition());
  cvShowImage("I2", pImgI2);
  
  //   .    
  CvPoint2D32f center = cvPoint2D32f(pImgI1->width/2, pImgI1->height/2);
  CvMat* pMatWarp = cvCreateMat(2, 3, CV_32F);
  cv2DRotationMatrix(center, fRotAngle, 1.0, pMatWarp);
  CV_MAT_ELEM(*pMatWarp, float, 0, 2) += fOffsX;
  CV_MAT_ELEM(*pMatWarp, float, 1, 2) += fOffsY;

  //    .
  float fAccuracy = CalcImageRegistrationAccuracy(
    rcROI, pMatWarp, m_PixelMapping.GetComposition());
  _tprintf(_T("Pixel match = %0.5g\n"), fAccuracy);

  //      .
  float fEstimatedRotAngle = 0;
  float fEstimatedOffsX = 0;
  float fEstimatedOffsY = 0;    
  m_PixelMapping.ExtractParams(center, fEstimatedRotAngle, fEstimatedOffsX, fEstimatedOffsY);
  
  //      .
  _tprintf(_T("True parameters: RotAngle = %0.5g OffsX = %0.5g OffsY = %0.5g\n"), fRotAngle, fOffsX, fOffsY);

  //     .
  _tprintf(_T("Estimated parameters: RotAngle = %0.5g OffsX = %0.5g OffsY = %0.5g\n"), 
      fEstimatedRotAngle, fEstimatedOffsX, fEstimatedOffsY);
  
  //  -.
  fclose(f);

  //   .
  cvWaitKey(0);

  // .    
  cvDestroyWindow("I1");
  cvDestroyWindow("I2");
  cvDestroyWindow("ConfMap");
  cvReleaseImage(&pImgI1);
  cvReleaseImage(&pImgConfMap);
  cvReleaseImage(&pImgI2);    
}

  
void CTest::RunPerformanceTest()
{
  //  .
  cvNamedWindow("I1");
  cvMoveWindow("I1", 10, 200);
  cvNamedWindow("I2");
  cvMoveWindow("I2", 400, 200);
  cvNamedWindow("ConfMap");
  cvMoveWindow("ConfMap", 10, 500);
 
  //    .
  CvRNG rng = cvRNG(time(NULL));
  //   100 x 1   .
  CvMat* pMatRandParams = cvCreateMat(100, 1, CV_32FC3);
  //     ,   .
  cvRandArr(&rng, pMatRandParams, CV_RAND_NORMAL, cvScalar(0), cvScalar(3, 5, 5));

  //    100        .
  //      ,       .
  int n;
  for(n = 1; n<=2; n++)
  {
    //     .
    CStatSample ImageRegTimeSample;
    //      .
    CStatSample ControlPointMatchAccuracySample;
    //    .
    CStatSample IterCountSample;

    //     .
    int nRegCount;
    for(nRegCount=0; nRegCount<100; nRegCount++)
    {
      //       .
      float fRotAngle = CV_MAT_ELEM(*pMatRandParams, float, nRegCount, 0);
      float fOffsX = CV_MAT_ELEM(*pMatRandParams, float, nRegCount, 0);
      float fOffsY = CV_MAT_ELEM(*pMatRandParams, float, nRegCount, 0);

      //      .
      IplImage* pImgI1 = NULL;
      IplImage* pImgConfMap = NULL;
      IplImage* pImgI2 = NULL;

      //   .
      CvRect rcROI = cvRect(100, 50, 150, 150);    
      if(!CreateTestImages("Koala.jpg", &rcROI, 
        fRotAngle, fOffsX, fOffsY, &pImgI1, &pImgConfMap, &pImgI2))
      {
        //   .
        return;
      }
  
      //       .
      m_PixelMapping.Reset();

      CTimeMeter tmeter;
      tmeter.Start();

      //   .
      int nIters = 0;
      bool bRegister = m_ImgRegAlg.RegisterImages(pImgI1, pImgConfMap, pImgI2, 
        &m_PixelMapping, 10E-03f, 100, 5, &nIters, 10E-03f, n);
      assert(bRegister);

      double dDuration = tmeter.Finish();
      ImageRegTimeSample.AddValue(dDuration);      
   
      IterCountSample.AddValue(nIters);

      //     I1.
      CvPoint TopLeft = cvPoint(rcROI.x, rcROI.y);
      CvPoint BottomRight = cvPoint(rcROI.x+rcROI.width, rcROI.y+rcROI.height);
      cvRectangle(pImgI1, TopLeft, BottomRight, CV_RGB(255, 255,255));
      cvShowImage("I1", pImgI1);

      //     I2.
      DrawWarpedRect(pImgI2, rcROI, m_PixelMapping.GetComposition());
      cvShowImage("I2", pImgI2);
      cvWaitKey(1);
      
      //   .    
      CvPoint2D32f center = cvPoint2D32f(pImgI1->width/2, pImgI1->height/2);
      CvMat* pMatWarp = cvCreateMat(2, 3, CV_32F);
      cv2DRotationMatrix(center, fRotAngle, 1.0, pMatWarp);
      CV_MAT_ELEM(*pMatWarp, float, 0, 2) += fOffsX;
      CV_MAT_ELEM(*pMatWarp, float, 1, 2) += fOffsY;

      //    .
      float fAccuracy = CalcImageRegistrationAccuracy(
        rcROI, pMatWarp, m_PixelMapping.GetComposition());
      ControlPointMatchAccuracySample.AddValue(fAccuracy);

      _tprintf(_T("Image registration experiment %d.\n"), nRegCount);
      _tprintf(_T("Control point match accuracy = %0.5g\n"), fAccuracy);

      //      .
      float fEstimatedRotAngle = 0;
      float fEstimatedOffsX = 0;
      float fEstimatedOffsY = 0;
      m_PixelMapping.ExtractParams(center, fEstimatedRotAngle, fEstimatedOffsX, fEstimatedOffsY);
      
      //      .
      _tprintf(_T("True parameters: RotAngle = %0.5g OffsX = %0.5g OffsY = %0.5g\n"), fRotAngle, fOffsX, fOffsY);

      //     .
      _tprintf(_T("Estimated parameters: RotAngle = %0.5g OffsX = %0.5g OffsY = %0.5g\n"), 
          fEstimatedRotAngle, fEstimatedOffsX, fEstimatedOffsY);

      _tprintf(_T("Time = %0.3g sec.\n"), dDuration);

      // .
      cvReleaseImage(&pImgI1);
      cvReleaseImage(&pImgConfMap);
      cvReleaseImage(&pImgI2);    
    }
  
    //    .
    TCHAR* szFileName = NULL;
    if(n==1)
      szFileName = _T("stat1.txt");
    else
      szFileName = _T("stat2.txt");
    FILE* f = NULL;
    _tfopen_s(&f, szFileName, _T("wt"));
    ImageRegTimeSample.PrintStat(f, "Image regisgration time");
    ControlPointMatchAccuracySample.PrintStat(f, "Control point matching accuracy");
    IterCountSample.PrintStat(f, "Iteration count");
    fclose(f);
  }

  // .    
  cvDestroyWindow("I1");
  cvDestroyWindow("I2");
  cvDestroyWindow("ConfMap");    
  cvReleaseMat(&pMatRandParams);
}

  
  
bool CTest::CreateTestImages(
    char* szFileName,     // [in]    .
    CvRect* prcROI,       // [in]   .
    float fRotAngle,      // [in]  .
    float fOffsX,         // [in]    OX.
    float fOffsY,         // [in]    OY.
    IplImage** ppImgI1,   // [out]  I1.
    IplImage** ppImgConfMap,  // [out]    I1.
    IplImage** ppImgI2    // [out]  I2.
    )
{
  //  .
  IplImage* pImgPhoto = cvLoadImage(szFileName);
  if(pImgPhoto==NULL)
  {
    printf("Error loading image %s.\n", szFileName);
    return false;
  }

  //    grayscale (  I1).
  CvSize ImageSize = cvSize(pImgPhoto->width, pImgPhoto->height);
  *ppImgI1 = cvCreateImage(ImageSize, IPL_DEPTH_8U, 1);  
  cvCvtColor(pImgPhoto, *ppImgI1, CV_RGB2GRAY);  
  //   .
  cvReleaseImage(&pImgPhoto);

  //   .    
  CvPoint2D32f center = cvPoint2D32f(ImageSize.width/2, ImageSize.height/2);
  CvMat* pMatWarp = cvCreateMat(2, 3, CV_32F);
  cv2DRotationMatrix(center, fRotAngle, 1.0, pMatWarp);
  CV_MAT_ELEM(*pMatWarp, float, 0, 2) += fOffsX;
  CV_MAT_ELEM(*pMatWarp, float, 1, 2) += fOffsY;

  //   .
  *ppImgConfMap = cvCreateImage(ImageSize, IPL_DEPTH_8U, 1);
  cvSet(*ppImgConfMap, cvScalar(0));  
  cvSetImageROI(*ppImgConfMap, *prcROI);
  cvSet(*ppImgConfMap, cvScalar(255));  
  cvResetImageROI(*ppImgConfMap);

  //   ,   .
  *ppImgI2 = cvCreateImage(ImageSize, IPL_DEPTH_8U, 1);
  cvWarpAffine(*ppImgI1, *ppImgI2, pMatWarp, CV_WARP_FILL_OUTLIERS);

  // 
  cvReleaseMat(&pMatWarp);

  return true;
}

float CTest::CalcImageRegistrationAccuracy(
  CvRect& rcROI,    // [in] ,      .
  CvMat* pMatWarp,  // [in]   .
  CvMat* pMatEstimatedWarp) // [in]  ,    .
{
  //       (   ).
  CvPoint2D32f ControlPoints[3];
  ControlPoints[0] = cvPoint2D32f(rcROI.x, rcROI.y);
  ControlPoints[1] = cvPoint2D32f(rcROI.x+rcROI.width, rcROI.y+rcROI.height);
  ControlPoints[2] = cvPoint2D32f(rcROI.x+rcROI.width, rcROI.y);

  float fMeanDistance = 0;
  int i;
  for(i=0; i<3; i++)
  {
    //       .
    CvPoint2D32f ptWarped = WarpPoint(&ControlPoints[i], pMatWarp);
    //      ,  
    //   .
    CvPoint2D32f ptWarped2 =  WarpPoint(&ControlPoints[i], pMatEstimatedWarp);

    //    .
    float fDiffX = (ptWarped2.x-ptWarped.x);
    float fDiffY = (ptWarped2.y-ptWarped.y);
    fMeanDistance += sqrt(fDiffX*fDiffX + fDiffY*fDiffY);
  }

  fMeanDistance /= 3;
  return fMeanDistance;    
}


CvPoint2D32f 
CTest::WarpPoint(
   CvPoint2D32f* pSrcPoint, 
   CvMat* pMatWarp)
{
  CvPoint2D32f p;
  
  p.x = CV_MAT_ELEM(*pMatWarp, float, 0, 0)*pSrcPoint->x + 
        CV_MAT_ELEM(*pMatWarp, float, 0, 1)*pSrcPoint->y + CV_MAT_ELEM(*pMatWarp, float, 0, 2);

  p.y = CV_MAT_ELEM(*pMatWarp, float, 1, 0)*pSrcPoint->x + 
        CV_MAT_ELEM(*pMatWarp, float, 1, 1)*pSrcPoint->y + CV_MAT_ELEM(*pMatWarp, float, 1, 2);

  return p;
}

void CTest::DrawWarpedRect(
  IplImage* pImg, // [in] ,   .
  CvRect& rc,     // [in] .
  CvMat* pMatWarp // [in]  .
  ) 
{
  CvPoint2D32f TopLeft = cvPoint2D32f(rc.x, rc.y);
  CvPoint2D32f TopRight = cvPoint2D32f(rc.x+rc.width, rc.y);
  CvPoint2D32f BottomLeft = cvPoint2D32f(rc.x, rc.y+rc.height);
  CvPoint2D32f BottomRight = cvPoint2D32f(rc.x+rc.width, rc.y+rc.height);

  //   .
  TopLeft = WarpPoint(&TopLeft, pMatWarp);
  TopRight = WarpPoint(&TopRight, pMatWarp);
  BottomLeft = WarpPoint(&BottomLeft, pMatWarp);
  BottomRight = WarpPoint(&BottomRight, pMatWarp);

  //  .

  CvScalar color = CV_RGB(255, 255, 255);

  cvLine(pImg, cvPoint(cvRound(TopLeft.x), cvRound(TopLeft.y)), 
    cvPoint(cvRound(TopRight.x), cvRound(TopRight.y)), color);

  cvLine(pImg, cvPoint(cvRound(TopRight.x), cvRound(TopRight.y)), 
    cvPoint(cvRound(BottomRight.x), cvRound(BottomRight.y)), color);

  cvLine(pImg, cvPoint(cvRound(BottomRight.x), cvRound(BottomRight.y)), 
    cvPoint(cvRound(BottomLeft.x), cvRound(BottomLeft.y)), color);

  cvLine(pImg, cvPoint(cvRound(BottomLeft.x), cvRound(BottomLeft.y)), 
    cvPoint(cvRound(TopLeft.x), cvRound(TopLeft.y)), color);
}


void CTest::RunCorrectnessTest()
{
  TestPixelMapping3DOF();
  TestInterpolatePixel();
  TestWarpImageI2();
  TestIsInnerLoopReady();
  TestIsReady();

  printf("Correctness test has finished.\n");
}

void CTest::TestPixelMapping3DOF()
{
  //    .
  
  CPixelMappingModel3DOF PixelMapping;
  CvPoint2D32f src_pixel = cvPoint2D32f(123.523f, 15.645f);
  
  // ,       .
  CvPoint2D32f dst_pixel0;
  bool bMap0 = PixelMapping.MapPixel(&src_pixel, &dst_pixel0);
  TEST_ASSERT(bMap0);

  //   .
  CvMat* pMatJacobian = cvCreateMat(2, 3, CV_32FC1);
  bool bCalc = PixelMapping.CalcJacobianMat(pMatJacobian, &src_pixel);
  TEST_ASSERT(bCalc);
  TEST_ASSERT(CV_MAT_ELEM(*pMatJacobian, float, 0, 0)==-src_pixel.y);
  cvReleaseMat(&pMatJacobian);

  //  .
  CvMat* pMatParams = cvCreateMat(3, 1, CV_32FC1);
  float wz = 0.186f;
  float tx = 3.895f;
  float ty = -2.531f;
  CV_MAT_ELEM(*pMatParams, float, 0, 0) = wz;
  CV_MAT_ELEM(*pMatParams, float, 1, 0) = tx;
  CV_MAT_ELEM(*pMatParams, float, 2, 0) = ty;
  bool bUpdate = PixelMapping.Update(pMatParams);
  TEST_ASSERT(bUpdate);
  cvReleaseMat(&pMatParams);
  
  //   .  
  CvPoint2D32f dst_pixel;
  bool bMap = PixelMapping.MapPixel(&src_pixel, &dst_pixel);
  CvMat* pMatComposition = PixelMapping.GetComposition();
  float c_11 = CV_MAT_ELEM(*pMatComposition, float, 0, 0);
  float c_12 = CV_MAT_ELEM(*pMatComposition, float, 0, 1);
  float c_13 = CV_MAT_ELEM(*pMatComposition, float, 0, 2);
  float c_21 = CV_MAT_ELEM(*pMatComposition, float, 1, 0);
  float c_22 = CV_MAT_ELEM(*pMatComposition, float, 1, 1);
  float c_23 = CV_MAT_ELEM(*pMatComposition, float, 1, 2);
  TEST_ASSERT(bMap);  
  float correct_c11 = 1.0f/(wz*wz+1.0f);
  float correct_c12 = wz/(wz*wz+1.0f);
  float correct_c13 = -tx*correct_c11 - ty*correct_c12;
  float correct_c21 = -wz/(wz*wz+1.0f);
  float correct_c22 = 1.0f/(wz*wz+1.0f);  
  float correct_c23 = -tx*correct_c21 - ty*correct_c22;
  float correct_x = correct_c11*src_pixel.x + correct_c12*src_pixel.y + correct_c13;
  float correct_y = correct_c21*src_pixel.x + correct_c22*src_pixel.y + correct_c23;  
  TEST_ASSERT(correct_x == dst_pixel.x && correct_y==dst_pixel.y);

  //    .
  bool bReset = PixelMapping.Reset();
  TEST_ASSERT(bReset);
  
  // ,      .
  CvPoint2D32f dst_pixel2;
  bool bMap2 = PixelMapping.MapPixel(&src_pixel, &dst_pixel2);
  TEST_ASSERT(bMap2);
  TEST_ASSERT(src_pixel.x==dst_pixel2.x && src_pixel.y==dst_pixel2.y);

  //    .
  float fRotAngle, fOffsX, fOffsY;
  PixelMapping.ExtractParams(cvPoint2D32f(0,0), fRotAngle, fOffsX, fOffsY);
  TEST_ASSERT(fRotAngle==0 && fOffsX==0 && fOffsY==0);
}

void CTest::TestInterpolatePixel()
{
  //    

  //   .
  CvSize image_size = cvSize(320, 240);
  IplImage* pImgTest = cvLoadImage("Koala.jpg", 0);
  TEST_ASSERT(pImgTest);

  CInvCompRegistrationAlgorithm RegAlg;

  // ,        .
  int i, j;
  for(i=0; i<pImgTest->width; i++)
  {
    for(j=0; j<pImgTest->height; j++)
    {      
      float fInterpolatedIntensity = RegAlg.InterpolatePixelIntensity(pImgTest, (float)i, (float)j);
      float fIntensity = (float)CV_IMAGE_ELEM(pImgTest, uchar, j, i);
      TEST_ASSERT(fIntensity==fInterpolatedIntensity);
    }
  }

  //  .
  cvReleaseImage(&pImgTest);
}

void CTest::TestWarpImageI2()
{
  //      I2.
  //   .
  CvSize image_size = cvSize(320, 240);
  IplImage* pImgTest = cvLoadImage("Koala.jpg", 0);
  TEST_ASSERT(pImgTest);

  IplImage* pImgConfMap = cvCreateImage(image_size, IPL_DEPTH_8U, 1);
  TEST_ASSERT(pImgConfMap);  
  cvSet(pImgConfMap, cvScalar(255));

  CInvCompRegistrationAlgorithm RegAlg;
  CPixelMappingModel3DOF PixelMapping;

  CvMat* pMatParams = cvCreateMat(3, 1, CV_32FC1);
  float wz = 0;
  float tx = 3.0f;
  float ty = -2.0f;
  CV_MAT_ELEM(*pMatParams, float, 0, 0) = wz;
  CV_MAT_ELEM(*pMatParams, float, 1, 0) = tx;
  CV_MAT_ELEM(*pMatParams, float, 2, 0) = ty;
  bool bUpdate = PixelMapping.Update(pMatParams);
  TEST_ASSERT(bUpdate);
  cvReleaseMat(&pMatParams);
  
  bool bAlloc = RegAlg.AllocResources(image_size, &PixelMapping);
  TEST_ASSERT(bAlloc);

  IplImage* pImgWarpedI2 = RegAlg.WarpImageI2(pImgTest, pImgConfMap, &PixelMapping);

  int i, j;
  for(i=0; i<pImgTest->width; i++)
  {
    for(j=0; j<pImgTest->height; j++)
    {
      int warped_i = i+(int)tx;
      int warped_j = j+(int)ty;
      if(warped_i>=0 && warped_i<pImgTest->width &&
         warped_j>=0 && warped_j<pImgTest->height)
      {
        uchar uWarpedIntensity = CV_IMAGE_ELEM(pImgWarpedI2, uchar, warped_j, warped_i);      
        uchar uCorrectIntensity = CV_IMAGE_ELEM(pImgTest, uchar, j, i);
        TEST_ASSERT(uCorrectIntensity==uWarpedIntensity);
      }
    }
  }

  //  .
  cvReleaseImage(&pImgTest);
}

void CTest::TestIsInnerLoopReady()
{
  //     .

  CInvCompRegistrationAlgorithm RegAlg;
  CRITERIA_TYPE crit = NONE;
  CvMat* pMatPrevParams = cvCreateMat(3, 1, CV_32FC1);
  CvMat* pMatNewParams = cvCreateMat(3, 1, CV_32FC1);
  
  CV_MAT_ELEM(*pMatPrevParams, float, 0, 0) = 0;
  CV_MAT_ELEM(*pMatPrevParams, float, 1, 0) = 0;
  CV_MAT_ELEM(*pMatPrevParams, float, 2, 0) = 0;

  CV_MAT_ELEM(*pMatNewParams, float, 0, 0) = -0.01f;
  CV_MAT_ELEM(*pMatNewParams, float, 1, 0) = 0.01f;
  CV_MAT_ELEM(*pMatNewParams, float, 2, 0) = -0.02f;

  //     .
  crit = RegAlg.IsInnerLoopReady(pMatPrevParams, pMatNewParams, 1E-03f, 98, 100);
  TEST_ASSERT(crit==NONE);

  //  .  .
  crit = RegAlg.IsInnerLoopReady(pMatPrevParams, pMatNewParams, 1E-03f, 99, 100);
  TEST_ASSERT(crit==MAX_ITERS_REACHED);

  //   .
  CV_MAT_ELEM(*pMatNewParams, float, 0, 0) = -0.0004f;
  CV_MAT_ELEM(*pMatNewParams, float, 1, 0) = -0.0005f;
  CV_MAT_ELEM(*pMatNewParams, float, 2, 0) = -0.0006f;
  crit = RegAlg.IsInnerLoopReady(pMatPrevParams, pMatNewParams, 1E-03f, 0, 100);
  TEST_ASSERT(crit==ACCURACY_REACHED);

  CV_MAT_ELEM(*pMatNewParams, float, 0, 0) = 0.0004f;
  CV_MAT_ELEM(*pMatNewParams, float, 1, 0) = 0.0005f;
  CV_MAT_ELEM(*pMatNewParams, float, 2, 0) = 0.0006f;
  crit = RegAlg.IsInnerLoopReady(pMatPrevParams, pMatNewParams, 1E-03f, 0, 100);
  TEST_ASSERT(crit==ACCURACY_REACHED);

  cvReleaseMat(&pMatPrevParams);
  cvReleaseMat(&pMatNewParams);
}

void CTest::TestIsReady()
{
  //     .

  CriteriaContext context;

  CInvCompRegistrationAlgorithm RegAlg;
  CRITERIA_TYPE crit = NONE;
  CvMat* pMatParams = cvCreateMat(3, 1, CV_32FC1);  
    
  CV_MAT_ELEM(*pMatParams, float, 0, 0) = -0.01f;
  CV_MAT_ELEM(*pMatParams, float, 1, 0) = 0.01f;
  CV_MAT_ELEM(*pMatParams, float, 2, 0) = -0.02f;

  //     .
  crit = RegAlg.IsReady(pMatParams, 1E-03f, 98, 100, 1E+06, 4, context);
  TEST_ASSERT(crit==NONE);

  //  .  .
  crit = RegAlg.IsReady(pMatParams, 1E-03f, 99, 100, 1E+06, 4, context);
  TEST_ASSERT(crit==MAX_ITERS_REACHED);

  //   .
  CV_MAT_ELEM(*pMatParams, float, 0, 0) = -0.0004f;
  CV_MAT_ELEM(*pMatParams, float, 1, 0) = -0.0005f;
  CV_MAT_ELEM(*pMatParams, float, 2, 0) = -0.0006f;
  crit = RegAlg.IsReady(pMatParams, 1E-03f, 0, 100, 1E+06, 4, context);  
  TEST_ASSERT(crit==ACCURACY_REACHED);

  CV_MAT_ELEM(*pMatParams, float, 0, 0) = 0.0004f;
  CV_MAT_ELEM(*pMatParams, float, 1, 0) = 0.0005f;
  CV_MAT_ELEM(*pMatParams, float, 2, 0) = 0.0006f;
  crit = RegAlg.IsReady(pMatParams, 1E-03f, 0, 100, 1E+06, 4, context);
  TEST_ASSERT(crit==ACCURACY_REACHED);

  //   
  crit = RegAlg.IsReady(pMatParams, 1E-05f, 0, 100, 1E+06, 4, context);
  TEST_ASSERT(crit==OSCILLATION);

  cvReleaseMat(&pMatParams); 
}